## PaCAL: A Python Package for Arithmetic Computations with Random Variables

from pylab import *
from pacal import *

def ishow():
    """A nonblocking matplotlib show"""
    ion()
    show()
    ioff()
def printhead(txt):
    """Pring a header"""
    print
    print txt
    print
if __name__ == "__main__":

    ###################################################################
    ###################################################################
    ## Section 1. Introduction
    ###################################################################
    ###################################################################

    ###############################################################
    printhead(""" Section 1.1. Basic usage """)
    ###############################################################
    Y = UniformDistr(1, 2)
    X = UniformDistr(3, 4)
    A = atan(Y / X)
    figure()
    A.plot()
    ishow()
    print A.mean()         # mean value of A
    print A.interval(0.95) # symmetric interval that supports 95% of the distribution                          
    A.summary()

    ###############################################################
    printhead(""" Section 1.5. A remark on semantics """)
    ###############################################################
    S1 = UniformDistr() + UniformDistr()
    U = UniformDistr(); S2 = U + U



    ###################################################################
    ###################################################################
    ## Section 2. Detailed description of PaCALs capabilities
    ###################################################################
    ###################################################################

    ###############################################################
    printhead(""" Section 2.3 Numerical accuracy.""")
    ###############################################################
    X1 = ChiSquareDistr(1)
    X2 = ChiSquareDistr(17)
    X3 = ChiSquareDistr(82)
    X4 = ChiSquareDistr(100)
    S = X1 + X2 + X3 + X4
    S.summary()

    X = NoncentralChiSquareDistr(df = 10, lmbda = 3)
    for k in range(11):
        print k, "m_k=", repr(X.moment(k, 0)), ",\tmu_k=", repr(X.moment(k))

    from scipy.stats import ncx2
    mu = sqrt(0.3)
    df = 10
    lmbda= df * mu ** 2.0
    S = NormalDistr(mu, 1) ** 2
    for i in range(df - 1):
        S += NormalDistr(mu, 1) ** 2
    X = NoncentralChiSquareDistr(df = 10, lmbda = 3)
    print "PaCAL simple                : ", repr(X.median())
    print "PaCAL direct                : ", repr(S.median())
    print "scipy.stats                 : ", repr(ncx2.median(df, lmbda))
    print "R: qchisq(0.5, df=10, ncp=3): ", repr(12.210896063727795)

    ###############################################################
    printhead(""" 2.4. Functions of random variables """)
    ###############################################################
    N = abs(NormalDistr()) ** NormalDistr()
    E = exp(CauchyDistr()) + exp(CauchyDistr())
    print N.int_error(), N.median()
    print E.int_error()
    figure()
    subplot(211)
    N.plot(xmin = 0, xmax = 3, label = "N")
    N.hist(xmin = 0, xmax = 3, bins = 100)
    axis((0, 3, 0, 2))
    legend()
    subplot(212)
    E.plot(xmin = 0, xmax = 4, label = "E")
    E.hist(xmin = 0, xmax = 4, bins = 100)
    axis((0, 4, 0, 0.5))
    legend()
    ishow()

    S = sign(NormalDistr()) * abs(NormalDistr())
    for i in range(11):
        print i, S.moment(i,0)


    ###############################################################
    printhead(""" 2.5. Random number generation """)
    ###############################################################
    import time
    X = NormalDistr()
    n = 1000000
    t0 = time.time(); A = X.rand(n); print time.time() - t0, "[s]"
    params.general.distr.independent = False
    t0 = time.time(); B = X.rand_invcdf(n); print time.time() - t0, "[s]"    # invcdf with interpolation
    params.general.distr.independent = True
    t0 = time.time();
    C = X.rand_invcdf(1000, use_interpolated = False); # invcdf without interpolation
    print time.time() - t0, "[s]"
    figure()
    subplot(311)
    hist(A, bins = 50, range = [-5, 5])
    subplot(312)
    hist(B, bins = 50, range = [-5, 5])
    subplot(313)
    hist(C, bins = 50, range = [-5, 5])
    ishow()

    ###############################################################
    printhead(""" 2.7. Operating directly on probability density functions and cumulative distribution functions """)
    ###############################################################
    U = UniformDistr()      # true distribution
    X = U.rand(50)          # a random sample of size 50
    D = DiscreteDistr(xi = X, pi = ones_like(X)/len(X)) # empirical distribution
    r = D.get_piecewise_cdf() - U.get_piecewise_cdf()   # pointwise difference
    figure()
    r.plot()
    ishow()
    x, D_n = r.max_abs()
    print "D_n =", D_n, ", K_n =", sqrt(50) * D_n


    ###############################################################
    printhead(""" 2.8.  Discrete and mixed distributions """)
    ###############################################################
    V, U = ZeroDistr(), ZeroDistr()
    for i in range(4):
        V += BernoulliDistr() / 4 ** (i + 1)    # Cantor distribution approximation
        U += 2 * BernoulliDistr() / 4 ** (i + 1)
    X = V + U
    UniformDistr().summary()
    X.summary()                              # Uniform distribution approximation
    figure()
    V.get_piecewise_cdf().plot(label = "V")
    axis((0, 0.34, 0, 1))
    legend()
    figure()
    X.get_piecewise_cdf().plot(linestyle = "--", label = "U + V")
    legend()
    ishow()

    NUM = UniformDistr(0, 1) + BinomialDistr(n = 2, p = 0.6)
    DEN = UniformDistr(-1, 0) + BinomialDistr(n = 2, p = 0.3)
    Q = NUM / DEN
    Q.summary()
    figure()
    Q.plot(xmin = -10, xmax = 10, linewidth = 2.0)
    Q.hist(xmin = -10, xmax = 10)
    ishow()

    D = DiscreteDistr(xi = [0.0, 0.2, 0.4, 0.6, 0.8], pi = [0.2, 0.2, 0.2, 0.2, 0.2])
    C = UniformDistr()
    M = min(C, D)
    S = M + UniformDistr()
    figure()
    M.plot(linewidth = 1.0, label = M.getName())
    S.plot(linestyle = '-', linewidth = 2.0, label = S.getName())
    legend()
    ishow()


    ###############################################################
    printhead(""" 2.9. Conditional distributions """)
    ###############################################################
    E = ExponentialDistr()
    CE = E | Gt(7)
    r = E.get_piecewise_pdf() - (CE - 7).get_piecewise_pdf()
    print "absolute_error =", r.max_abs()[1]


    ###############################################################
    printhead(""" 2.10. Operations on i.i.d. random variables """)
    ###############################################################
    X = ExponentialDistr()
    figure()
    GumbelDistr().plot(linestyle = "--", color = "r", linewidth = 3)
    for n in range(4, 25, 5):
        M = iid_max(X, n)
        (M - log(n)).plot()
    GumbelDistr().summary()
    (M - log(n)).summary()
    ishow()


    ###############################################################
    printhead(""" 2.11. Distribution fitting """)
    ###############################################################
    sample = BetaDistr(2, 4).rand(100)
    MLE = LoglikelihoodEstimator(BetaDistr, xi = sample)
    par = MLE.find_params()
    print par               # print the fitted parameters
    B = BetaDistr(**par)    # reconstruct the distribution with given parameters
    BetaDistr().summary()
    B.summary()



    ###################################################################
    ###################################################################
    ## Section 3. Towards dependent random variables
    ###################################################################
    ###################################################################

    ###############################################################
    printhead(""" Section 3.2. Examples: Copulas, sum of uniforms """)
    ###############################################################
    X = UniformDistr(0, 1, sym = "X")
    Y = UniformDistr(0, 1, sym = "Y")
    for theta, ls in [(-15, "--"), (-5, "-."), (0.05, "-"), (10, ":")]:
        Copula = FrankCopula(marginals = [X, Y], theta = theta)
        M = TwoVarsModel(Copula, X + Y)
        figure()
        Copula.plot()
        S = M.eval()
        figure()
        S.plot(linestyle = ls)
        S.summary()
        ishow()


    ###############################################################
    printhead(""" Section 3.2. Examples: Parallel circuit of two resistors. """)
    ###############################################################
    R1 = UniformDistr(0.5, 1.5)
    R2 = UniformDistr(1.5, 2.5)
    pr = PiCopula(marginals = [R1, R2])
    M = TwoVarsModel(pr, R1 * R2 / (R1 + R2))
    R_equiv = M.eval()
    figure()
    R_equiv.plot(linewidth = 2.0, color = 'k')
    R_equiv.hist()
    R_equiv.summary()
    ishow()

    from functools import partial
    ###############################################################
    printhead(""" Section 3.2. Examples: General regression. """)
    ###############################################################
    def funy(F, X, x, y):
        if isscalar(y):
            return F.pdf(x, y) / X.pdf(x)
        else:
            return array([F.pdf(x, yi) / X.pdf(x) for yi in y])

    def condmean(x, F, X, Y):
        #fun = lambda y: F.pdf(x,y) / X.pdf(x)     # conditional distribution
        bpt = Y.get_piecewise_pdf().getBreaks()
        distr = FunDistr(fun = partial(funy, F, X, x), breakPoints = bpt)
        return distr.mean()    # or median(), mode(), etc.

    X = UniformDistr(0, 1) + UniformDistr(0, 1) # triangular distribution
    X.setSym("X")
    Y = BetaDistr(1, 4, sym = "Y")
    F = FrankCopula(theta = 4, marginals = [X, Y]) # joint distribution
    figure()
    F.contour()
    xx, yy = linspace(0.01, 1.99, 30), []
    for x in xx:
        yy.append(condmean(x, F, X, Y))
    plot(xx, yy)
    ishow()


    ###################################################################
    ###################################################################
    ## Section 2. Applications
    ###################################################################
    ###################################################################

    ###############################################################
    printhead(""" Section 4. Sample distribution of the difference of proportions. """)
    ###############################################################
    B1 = BinomialDistr(8, p = 4.0/8.0)
    B2 = BinomialDistr(7, p = 2.0/7.0)
    D = B1 / 8 - B2 / 7
    D.summary()
    figure()
    subplot(211)
    D.plot()
    subplot(212)
    D.get_piecewise_cdf().plot()
    print D.interval(0.05)
    ishow()

    ###############################################################
    printhead(""" Section 4. Hill's tail exponent estimator.""")
    ###############################################################
    def Hills_estimator(n, alpha, xmin = 1):
        d = ParetoDistr(alpha) | Gt(xmin)
        s = iid_sum(log(d / xmin), n)
        return 1 + n / s
    H = Hills_estimator(21, 0.8)
    print "mean = ", H.mean()
    print "mean = ", H.median()
    print "c.i. = ", H.interval()


    ###############################################################
    printhead(""" Section 4. Weighted sum of t distributions. """)
    ###############################################################
    theta = pi/4
    weightedT = sin(theta) * StudentTDistr(10) - cos(theta) * StudentTDistr(20)
    weightedT.summary()
    figure()
    weightedT.plot(xmin = -5, xmax = 5)
    ishow()


    ###############################################################
    printhead(""" Section 4. Sample range and interquantile range distribution. """)
    ###############################################################
    X = NormalDistr(0, 1, sym = "X")
    #OIJ = IJthOrderStatsNDDistr(X, n = 5, i = 1, j = 5)  # joint distribution for SAS/QC example
    OIJ = IJthOrderStatsNDDistr(X, n = 8, i = 2, j = 7)   # joint distribution
    X2, X7 = OIJ.marginals                          # marginal distributions
    M = TwoVarsModel(OIJ, X7 - X2)                  # model with given expression
    R = M.eval()                                    # change of variable and integration
    figure()
    OIJ.plot()
    figure()
    R.plot(linewidth = 3.0, label = "R=X7-X2");
    R.summary()
    X2.plot(linewidth = 2.0, label = "X2")
    X7.plot(label = "X7")
    legend()
    print (X7.mean() - X2.mean()) - R.mean()
    print repr(R.var()), repr(R.std())

    show()
